from datetime import datetime, timedelta
import pathlib
import tempfile
import logging
from time import time
from typing import Union
import os
import glob
import multiprocessing as mp

#from appdirs import user_data_dir
import boto3
from botocore import UNSIGNED
from botocore.config import Config
import numpy as np
import pandas as pd
import xarray as xr

from pyschism.dates import nearest_cycle

logger = logging.getLogger(__name__)


class AWSGrib2Inventory:

    def __init__(
            self,
            start_date: datetime = None,
            record = 2,
            pscr = None, #tmpdir to save grib files
            product='conus',
    ):
        self.start_date = nearest_cycle() if start_date is None else start_date
        self.pscr = pscr
        self.product = product

        self.forecast_cycle = self.start_date #nearest_cycle()

        # paginator=self.s3.get_paginator('list_objects_v2')
        # pages=paginator.paginate(Bucket=self.bucket, 
        #         Prefix=f'hrrr.{self.forecast_cycle.strftime("%Y%m%d")}'
        #                f'/{self.product}/')
        # data=[]
        # for page in pages:
        #     for obj in page['Contents']:
        #         data.append(obj) 

        # self.cycle=self.forecast_cycle.hour
        # tz='t{:02d}z'.format(self.cycle)
        # self.file_metadata = list(sorted([
        #     _['Key'] for _ in data if 'wrfsfcf' in _['Key'] and tz in _['Key'] and not 'idx' in _['Key']
        # ]))
        self.get_file_namelist(self.forecast_cycle)

        while (not self.file_metadata): 
            logger.info(f'No data for cycle t{self.forecast_cycle.hour:02d}z on {self.forecast_cycle}, try next cycle: +1!')
            self.forecast_cycle += timedelta(hours=1)
            self.get_file_namelist(self.forecast_cycle)

        for key in self.file_metadata[1:record*24+1]:
            filename = pathlib.Path(self.tmpdir) / key
            filename.parent.mkdir(parents=True, exist_ok=True)

            with open(filename, 'wb') as f:
                logger.info(f'Downloading file {key}, ')
                try:
                    self.s3.download_fileobj(self.bucket, key, f)
                except:
                    logger.info(f'file {key} is not available')

    def get_file_namelist(self, requested_date):
        paginator=self.s3.get_paginator('list_objects_v2')
        pages=paginator.paginate(Bucket=self.bucket, 
                Prefix=f'hrrr.{requested_date.strftime("%Y%m%d")}'
                       f'/{self.product}/')
        data=[]
        for page in pages:
            for obj in page['Contents']:
                data.append(obj) 

        self.cycle = requested_date.hour
        tz='t{:02d}z'.format(self.cycle)
        self.file_metadata = list(sorted([
            _['Key'] for _ in data if 'wrfsfcf' in _['Key'] and tz in _['Key'] and not 'idx' in _['Key']
        ]))
        return self.file_metadata

    @property
    def bucket(self):
        return 'noaa-hrrr-bdp-pds'

    @property
    def output_interval(self) -> timedelta:
        return {
            'conus': timedelta(hours=1)
        }[self.product]

    @property
    def s3(self):
        try:
            return self._s3
        except AttributeError:
            self._s3 = boto3.client(
                's3', config=Config(signature_version=UNSIGNED))
            return self._s3

    @property
    def tmpdir(self):
        if not hasattr(self, "_tmpdir"):
            self._tmpdir = tempfile.TemporaryDirectory(dir=self.pscr)
        return pathlib.Path(self._tmpdir.name)

    @property
    def files(self):
        grbfiles=glob.glob(f'{self.tmpdir}/hrrr.{self.forecast_cycle.strftime("%Y%m%d")}/conus/hrrr.t{self.cycle:02d}z.wrfsfcf*.grib2')
        grbfiles.sort()
        return grbfiles

class HRRR:

    def __init__(self, start_date=None, rnday=None, pscr=None, record=2, bbox=None, outdir=None):

        start_date = nearest_cycle() if start_date is None else start_date 
        self.bbox = bbox

        if outdir is None:
            self.path = pathlib.Path(start_date.strftime("%Y%m%d"))
            self.path.mkdir(parents=True, exist_ok=True)
        else:
            self.path = outdir

        end_date = start_date + timedelta(days=rnday)
        logger.info(f'start_date is {start_date}, end_date is {end_date}')
        
        datevector = np.arange(start_date, end_date, np.timedelta64(1, 'D'),
                dtype='datetime64')
        datevector = pd.to_datetime(datevector)

        #Use all CPUs may cause memory issue
        npool = len(datevector) if len(datevector) < mp.cpu_count()/2 else mp.cpu_count()/2
        logger.info(f'npool is {npool}')
        pool = mp.Pool(int(npool))

        pool.starmap(self.gen_sflux, [(date, record, pscr) for date in datevector])
        #self.gen_sflux(datevector[0], record, pscr)

        pool.close()
        
    def gen_sflux(self, date, record, pscr):

        inventory = AWSGrib2Inventory(date, record, pscr)
        grbfiles = inventory.files
        cycle = date.hour

        stmp = [] 
        spfh = []
        uwind = []
        vwind = []
        prmsl = []
        prate = [] 
        dlwrf = [] 
        dswrf = [] 

        Vars = {
            'group1': {'sh2': ['174096', spfh], 't2m': ['167', stmp],'u10': ['165', uwind], 'v10': ['166', vwind]},
            'group2': {'mslma': ['meanSea', prmsl]},
            'group3': {'prate': ['surface', prate], 'dlwrf': ['surface', dlwrf], 'dswrf': ['surface', dswrf]}
        }

        #Get lon/lat
        lon, lat, idx_ymin, idx_ymax, idx_xmin, idx_xmax = self.modified_latlon(grbfiles[0])
 

        times = []
        for ifile, fname in enumerate(grbfiles):
            logger.info(f'file {ifile} is {fname}')

            ds=xr.open_dataset(
                fname,
                engine='cfgrib',
                backend_kwargs=dict(filter_by_keys={'stepType': 'instant','typeOfLevel': 'surface'})
            )
            times.append(ds.valid_time.values)
            ds.close()

            for key, value in Vars.items():
                if key == 'group1':
                    for key2, value2 in value.items():
                        ds=xr.open_dataset(
                            fname,
                            engine='cfgrib',
                            backend_kwargs=dict(filter_by_keys={'paramId':int(value2[0])})
                        )
                        value2[1].append(ds[key2][idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32'))
                        ds.close()

                elif key == 'group2':
                    ds=xr.open_dataset(
                        fname,
                        engine='cfgrib',
                        backend_kwargs=dict(filter_by_keys={'typeOfLevel':'meanSea'})
                    )
                    for key2, value2 in value.items():
                        value2[1].append(ds[key2][idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32'))
                    ds.close()

                else:
                    ds=xr.open_dataset(
                        fname,
                        engine='cfgrib',
                        backend_kwargs=dict(filter_by_keys={'stepType': 'instant','typeOfLevel': 'surface'})
                    )
                    for key2, value2 in value.items():
                        value2[1].append(ds[key2][idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32'))
                    ds.close()    

        #write netcdf
        fout = xr.Dataset({
            'stmp': (['time', 'ny_grid', 'nx_grid'], np.array(stmp)),
            'spfh': (['time', 'ny_grid', 'nx_grid'], np.array(spfh)),
            'uwind': (['time', 'ny_grid', 'nx_grid'], np.array(uwind)),
            'vwind': (['time', 'ny_grid', 'nx_grid'], np.array(vwind)),
            'prmsl': (['time', 'ny_grid', 'nx_grid'], np.array(prmsl)),
            'prate': (['time', 'ny_grid', 'nx_grid'], np.array(prate)),
            'dlwrf': (['time', 'ny_grid', 'nx_grid'], np.array(dlwrf)),
            'dswrf': (['time', 'ny_grid', 'nx_grid'], np.array(dswrf)),
            },
            coords={
                'time': np.round((times - date.to_datetime64()) / np.timedelta64(1, 'D'), 5).astype('float32'),
                'lon': (['ny_grid', 'nx_grid'], lon),
                'lat': (['ny_grid', 'nx_grid'], lat)})

        #date_string = np.datetime_as_string(times[0], unit='m')
        #bdate = datetime.strptime(date_string, '%Y-%m-%dT%H:%M')
        bdate = date.strftime('%Y %m %d %H').split(' ')
        bdate = [int(q) for q in bdate[:4]] + [0]

        fout.time.attrs = {
            'long_name': 'Time',
            'standard_name': 'time',
            'base_date': bdate,
            'units': f'days since {date.year}-{date.month}-{date.day} {cycle:02d}:00 UTC'
        }

        fout.lat.attrs = {
            'units': 'degrees_north',
            'long_name': 'Latitude',
            'standard_name': 'latitude',
        }

        fout.lon.attrs = {
            'units': 'degrees_east',
            'long_name': 'Longitude',
            'standard_name': 'longitude',
        }

        fout.uwind.attrs={
            'units': 'm/s',
            'long_name': '10m_above_ground/UGRD',
            'standard_name':'eastward_wind'
        }
      
        fout.vwind.attrs={
            'units': 'm/s',
            'long_name': '10m_above_ground/UGRD',
            'standard_name':'northward_wind'
        }

        fout.spfh.attrs={
            'units': 'kg kg-1',
            'long_name': '2m_above_ground/Specific Humidity',
            'standard_name':'specific_humidity'
        }

        fout.prmsl.attrs = {
            'units': 'Pa',
            'long_name': 'Pressure reduced to MSL',
            'standard_name': 'air_pressure_at_sea_level'
        }

        fout.stmp.attrs={
            'units': 'K',
            'long_name': '2m_above_ground/Temperature',
        }

        fout.prate.attrs={
            'units': 'kg m-2 s-1',
            'long_name': 'Precipitation rate'
        }
        fout.dlwrf.attrs = {
            'units': 'W m-2',
            'long_name': 'Downward short-wave radiation flux'
        }

        fout.dswrf.attrs = {
            'units': 'W m-2',
            'long_name': 'Downward long-wave radiation flux'
        }
                         
        fout.to_netcdf(self.path / f'hrrr_{date.strftime("%Y%m%d")}{cycle:02d}.nc','w', 'NETCDF3_CLASSIC', unlimited_dims='time')

    def modified_latlon(self, grbfile):

        xmin, xmax, ymin, ymax = self.bbox.xmin, self.bbox.xmax, self.bbox.ymin, self.bbox.ymax
        xmin = xmin + 360 if xmin < 0 else xmin
        xmax = xmax + 360 if xmax < 0 else xmax
       
        ds=xr.open_dataset(grbfile,
            engine='cfgrib',
            backend_kwargs=dict(filter_by_keys={'stepType': 'instant','typeOfLevel': 'surface'}))
        lon=ds.longitude.astype('float32')
        lon_idxs=(lon.values >= xmin) & (lon.values <= xmax)
        lat=ds.latitude.astype('float32')
        lat_idxs=(lat.values >= ymin) & (lat.values <= ymax + 2.0)
        idxs = lon_idxs & lat_idxs
        idxs = np.argwhere(idxs)
        idx_ymin = np.min(idxs[:,0])
        idx_ymax = np.max(idxs[:,0])
        idx_xmin = np.min(idxs[:,1])
        idx_xmax = np.max(idxs[:,1])
     
        lon2 = lon.values[idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1]
        lat2 = lat.values[idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1]
        idxs = np.where(lon2 > 180)
        lon2[idxs] -= 360
        logger.info(f'idx_ymin is {idx_ymin}, idx_ymax is {idx_ymax}, idx_xmin is {idx_xmin}, idx_xmax is {idx_xmax}')
        ds.close()

        return lon2, lat2, idx_ymin, idx_ymax, idx_xmin, idx_xmax



#!/usr/bin/env python
import pathlib
import setuptools  # type: ignore[import]
import subprocess
import sys

subprocess.check_call(
    [sys.executable, '-m', 'pip', 'install', '--upgrade', 'pip'])

subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'wheel'])

try:
    from dunamai import Version
except ImportError:
    subprocess.check_call(
            [sys.executable, '-m', 'pip', 'install', 'dunamai']
    )
    from dunamai import Version  # type: ignore[import]

try:
    version = Version.from_any_vcs().serialize()
except RuntimeError:
    version = '0.0.0'
except ValueError as e:
    if "time data '%cI' does not match format '%Y-%m-%dT%H:%M:%S%z'" in str(e):
        version = '0.0.0'
    else:
        raise

class BuildSchism(setuptools.Command):

    description = "build external SCHISM dependencies"

    user_options = [
        ('url=', None, 'Path for git clone of SCHISM source.'),
        ('branch=', None, 'Branch to use for install'),
        ('hydro=', None, 'Branch to use for install')
    ]

    def initialize_options(self):
        self.url = None
        self.branch = None
        self.hydro = None

    def finalize_options(self):
        self.url = 'https://github.com/schism-dev/schism' if self.url is None \
            else self.url
        self.branch = 'master' if self.branch is None else self.branch
        self.hydro = True if self.hydro is None else bool(self.hydro)

    def run(self):
        print(self.url, self.branch, self.hydro)
        # subprocess.check_call(["git", "clone", "submodules/jigsaw-python"])
        # # install jigsawpy
        # os.chdir(PARENT / 'submodules/jigsaw-python')
        # subprocess.check_call(["git", "checkout", "master"])
        # self.announce('INSTALLING JIGSAWPY', level=3)
        # subprocess.check_call(["python", "setup.py", "install"])
        # # install jigsaw
        # self.announce(
        #     'INSTALLING JIGSAW LIBRARY AND BINARIES FROM '
        #     'https://github.com/dengwirda/jigsaw-python', level=3)
        # os.chdir("external/jigsaw")
        # os.makedirs("build", exist_ok=True)
        # os.chdir("build")
        # gcc, cpp = self._check_gcc_version()
        # subprocess.check_call(
        #     ["cmake", "..",
        #      "-DCMAKE_BUILD_TYPE=Release",
        #      f"-DCMAKE_INSTALL_PREFIX={PYENV_PREFIX}",
        #      f"-DCMAKE_C_COMPILER={gcc}",
        #      f"-DCMAKE_CXX_COMPILER={cpp}",
        #      ])
        # subprocess.check_call(["make", f"-j{cpu_count()}", "install"])
        # libsaw_prefix = list(PYENV_PREFIX.glob("**/*jigsawpy*")).pop() / '_lib'
        # os.makedirs(libsaw_prefix, exist_ok=True)
        # envlib = PYENV_PREFIX / 'lib' / SYSLIB[platform.system()]
        # os.symlink(envlib, libsaw_prefix / envlib.name)
        # os.chdir(PARENT)
        # subprocess.check_call(
        #   ["git", "submodule", "deinit", "-f", "submodules/jigsaw-python"])

parent = pathlib.Path(__file__).parent.absolute()
conf = setuptools.config.read_configuration(parent / 'setup.cfg')
meta = conf['metadata']
setuptools.setup(
    name=meta['name'],
    version=version,
    author=meta['author'],
    author_email=meta['author_email'],
    description=meta['description'],
    long_description=meta['long_description'],
    long_description_content_type="text/markdown",
    url=meta['url'],
    packages=setuptools.find_packages(exclude=['tests', 'examples', 'docs', 'docker']),
    python_requires='>3.6',
    setup_requires=['wheel', 'setuptools_scm', 'setuptools>=41.2',
                    'netcdf-flattener>=1.2.0'],
    include_package_data=True,
    extras_require={'dev': ['coverage', 'flake8', 'nose']},
    cmdclass={
        "build_schism": BuildSchism
    },
    install_requires=[
        'boto3',
        'cdsapi',
        'cf-python',
        'cfgrib',
        'f90nml',
        'fsspec',
        'geopandas',
        'metpy',
        'netCDF4',
        'netcdf-flattener>=1.2.0',
        'numba',
        'ordered-set',
        'psutil',
        'pygeos',
        'pyugrid',
        'rtree',
        'scipy',
        'seawater',
        'sqlalchemy',
        'stormevents',
        'tqdm',
        'tqdm-logging-wrapper',
        'utm',
        'wget',
        'xarray',
        'xmltodict',
        'zarr',
        'appdirs',
        'pylib-essentials',
    ],
    entry_points={'console_scripts': ['pyschism = pyschism.__main__:main']},
    tests_require=['nose'],
    test_suite='nose.collector',
)



from datetime import datetime, timedelta
import tempfile
import pathlib
from typing import Union
import logging

import appdirs
import numpy as np
import cdsapi
import netCDF4 as nc4
from netCDF4 import Dataset
import pandas as pd
from matplotlib.transforms import Bbox

from pyschism.forcing.nws.nws2.sflux import (
    SfluxDataset,
    AirComponent,
    PrcComponent,
    RadComponent,
)

logger = logging.getLogger(__name__)

class ERA5DataInventory:

    def __init__(self, start_date=None, rnday: Union[float, timedelta] = 4, bbox=None, tmpdir=None):

        self.start_date = start_date
        self.rnday = rnday
        self.end_date = self.start_date+timedelta(self.rnday + 1)
        self.client=cdsapi.Client()
        self._bbox = bbox
        if tmpdir is not None:
            self.tmpdir = tmpdir

        filename = self.tmpdir / f"era5_{self.start_date.strftime('%Y%m%d')}.nc"

        if filename.is_file() == False:

            r = self.client.retrieve(
                'reanalysis-era5-single-levels',
                {
                'variable':[
                    '10m_u_component_of_wind','10m_v_component_of_wind','mean_sea_level_pressure',
                    '2m_dewpoint_temperature','2m_temperature','mean_total_precipitation_rate',
                    'mean_surface_downward_long_wave_radiation_flux','mean_surface_downward_short_wave_radiation_flux'
                    ],
                'product_type':'reanalysis',
                'date':f"{self.start_date.strftime('%Y-%m-%d')}/{self.end_date.strftime('%Y-%m-%d')}",
                'time':[
                    '00:00','01:00','02:00','03:00','04:00','05:00',
                    '06:00','07:00','08:00','09:00','10:00','11:00',
                    '12:00','13:00','14:00','15:00','16:00','17:00',
                    '18:00','19:00','20:00','21:00','22:00','23:00'
                    ],
                'area': [self._bbox.ymax+0.5, self._bbox.xmin-0.5, self._bbox.ymin-0.5, self._bbox.xmax+0.5], # North, West, South, East. Default: global
                'format':'netcdf'
                }
            )
 
            r.download(filename)
        
    @property
    def tmpdir(self):
        if not hasattr(self, '_tmpdir'):
            self._tmpdir = tempfile.TemporaryDirectory()
        return pathlib.Path(self._tmpdir.name)

    @tmpdir.setter
    def tmpdir(self, value):
        self._tmpdir = value

    @property
    def files(self):
        return sorted(list(self.tmpdir.glob('**/era5_*.nc')))

    @property
    def lon(self):
        if not hasattr(self, '_lon'):
            self._lon = Dataset(self.files[0]).variables['longitude'][:]
            if not hasattr(self, '_lat'):
                self._lat = Dataset(self.files[0]).variables['latitude'][::-1]
        return self._lon

    @property
    def lat(self):
        if not hasattr(self, '_lat'):
            self._lat = Dataset(self.files[0]).variables['latitude'][::-1]
            if not hasattr(self, '_lon'):
                self._lon = Dataset(self.files[0]).variables['longitude'][:]
        return self._lat

    def xy_grid(self):
        return np.meshgrid(self.lon, self.lat)

def put_sflux_fields(iday, date, timevector, ds, nx_grid, ny_grid, air, rad, prc, output_interval, OUTDIR):
    rt=pd.to_datetime(str(date))
    idx=np.where(rt == timevector)[0].item()
    times=[i/24 for i in np.arange(0, 25, output_interval)]

    if air is True:
        with Dataset(OUTDIR / "sflux_air_1.{:04}.nc".format(iday+1), 'w', format='NETCDF3_CLASSIC') as dst:
            dst.setncatts({"Conventions": "CF-1.0"})
            # dimensions
            dst.createDimension('nx_grid', nx_grid.shape[1])
            dst.createDimension('ny_grid', ny_grid.shape[0])
            dst.createDimension('time', None)
            # variables
            # lon
            dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
            dst['lon'].long_name = "Longitude"
            dst['lon'].standard_name = "longitude"
            dst['lon'].units = "degrees_east"
            dst['lon'][:] = nx_grid
            # lat
            dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
            dst['lat'].long_name = "Latitude"
            dst['lat'].standard_name = "latitude"
            dst['lat'].units = "degrees_north"
            dst['lat'][:] = ny_grid
            # time
            dst.createVariable('time', 'f4', ('time',))
            dst['time'].long_name = 'Time'
            dst['time'].standard_name = 'time'
            dst['time'].units = f'days since {rt.year}-{rt.month}'\
                                f'-{rt.day} 00:00 UTC'
            dst['time'].base_date = (rt.year, rt.month, rt.day, 0)
            dst['time'][:] = times

            # prmsl
            dst.createVariable('prmsl', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['prmsl'].long_name = "Pressure reduced to MSL"
            dst['prmsl'].standard_name = "air_pressure_at_sea_level"
            dst['prmsl'].units = "Pa"
            dst['prmsl'][:,:,:]=ds['msl'][idx:idx+25:output_interval, ::-1, :]

            # spfh
            dst.createVariable('spfh', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['spfh'].long_name = "Surface Specific Humidity "\
                                    "(2m AGL)"
            dst['spfh'].standard_name = "specific_humidity"
            dst['spfh'].units = "1"
            #convert dewpoint to specific humidity
            d2m = ds['d2m'][idx:idx+25:output_interval, ::-1, :]
            msl = ds['msl'][idx:idx+25:output_interval, ::-1, :]
            Td = d2m - 273.15
            e1 = 6.112*np.exp((17.67*Td)/(Td + 243.5))
            spfh = (0.622*e1)/(msl*0.01 - (0.378*e1))
            dst['spfh'][:,:,:]=spfh

            # stmp
            dst.createVariable('stmp', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['stmp'].long_name = "Surface Air Temperature (2m AGL)"
            dst['stmp'].standard_name = "air_temperature"
            dst['stmp'].units = "K"
            dst['stmp'][:,:,:]=ds['t2m'][idx:idx+25:output_interval, ::-1, :]

            # uwind
            dst.createVariable('uwind', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['uwind'].long_name = "Surface Eastward Air Velocity "\
                "(10m AGL)"
            dst['uwind'].standard_name = "eastward_wind"
            dst['uwind'].units = "m/s"
            dst['uwind'][:,:,:]=ds['u10'][idx:idx+25:output_interval, ::-1, :]

            # vwind
            dst.createVariable('vwind', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['vwind'].long_name = "Surface Northward Air Velocity "\
                "(10m AGL)"
            dst['vwind'].standard_name = "northward_wind"
            dst['vwind'].units = "m/s"
            dst['vwind'][:,:,:]=ds['v10'][idx:idx+25:output_interval, ::-1, :]

    if prc is True:
        with Dataset(OUTDIR / "sflux_prc_1.{:04}.nc".format(iday+1), 'w', format='NETCDF3_CLASSIC') as dst:
            dst.setncatts({"Conventions": "CF-1.0"})
            # dimensions
            dst.createDimension('nx_grid', nx_grid.shape[1])
            dst.createDimension('ny_grid', ny_grid.shape[0])
            dst.createDimension('time', None)
            # variables
            # lon
            dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
            dst['lon'].long_name = "Longitude"
            dst['lon'].standard_name = "longitude"
            dst['lon'].units = "degrees_east"
            dst['lon'][:] = nx_grid
            # lat
            dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
            dst['lat'].long_name = "Latitude"
            dst['lat'].standard_name = "latitude"
            dst['lat'].units = "degrees_north"
            dst['lat'][:] = ny_grid
            # time
            dst.createVariable('time', 'f4', ('time',))
            dst['time'].long_name = 'Time'
            dst['time'].standard_name = 'time'
            dst['time'].units = f'days since {rt.year}-{rt.month}'\
                                f'-{rt.day} 00:00 UTC'
            dst['time'].base_date = (rt.year, rt.month, rt.day, 0)
            dst['time'][:] = times

            # prate
            dst.createVariable('prate', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['prate'].long_name = "Surface Precipitation Rate"
            dst['prate'].standard_name = "air_pressure_at_sea_level"
            dst['prate'].units = "kg/m^2/s"
            dst['prate'][:,:,:]=ds['mtpr'][idx:idx+25:output_interval, ::-1, :]

    if rad is True:
        with Dataset(OUTDIR / "sflux_rad_1.{:04}.nc".format(iday+1), 'w', format='NETCDF3_CLASSIC') as dst:
            dst.setncatts({"Conventions": "CF-1.0"})
            # dimensions
            dst.createDimension('nx_grid', nx_grid.shape[1])
            dst.createDimension('ny_grid', ny_grid.shape[0])
            dst.createDimension('time', None)
            # variables
            # lon
            dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
            dst['lon'].long_name = "Longitude"
            dst['lon'].standard_name = "longitude"
            dst['lon'].units = "degrees_east"
            dst['lon'][:] = nx_grid
            # lat
            dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
            dst['lat'].long_name = "Latitude"
            dst['lat'].standard_name = "latitude"
            dst['lat'].units = "degrees_north"
            dst['lat'][:] = ny_grid
            # time
            dst.createVariable('time', 'f4', ('time',))
            dst['time'].long_name = 'Time'
            dst['time'].standard_name = 'time'
            dst['time'].units = f'days since {rt.year}-{rt.month}'\
                                f'-{rt.day} 00:00 UTC'
            dst['time'].base_date = (rt.year, rt.month, rt.day, 0)
            dst['time'][:] = times

            # dlwrf
            dst.createVariable('dlwrf', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['dlwrf'].long_name = "Downward Long Wave Radiation Flux"
            dst['dlwrf'].standard_name = "surface_downwelling_longwave_flux_in_air"
            dst['dlwrf'].units = "W/m^2"
            dst['dlwrf'][:,:,:]=ds['msdwlwrf'][idx:idx+25:output_interval, ::-1, :]

            # dwrf
            dst.createVariable('dswrf', 'f4', ('time', 'ny_grid', 'nx_grid'))
            dst['dswrf'].long_name = "Downward Long Wave Radiation Flux"
            dst['dswrf'].standard_name = "surface_downwelling_shortwave_flux_in_air"
            dst['dswrf'].units = "W/m^2"
            dst['dswrf'][:,:,:]=ds['msdwswrf'][idx:idx+25:output_interval, ::-1, :]


class ERA5(SfluxDataset):

    def __init__(self, **kwargs):

        self.prmsl_name = 'prmslmsl'
        self.spfh_name = 'spfh2m'
        self.stmp_name = 'tmpsfc'
        self.uwind_name = 'ugrd10m'
        self.vwind_name = 'vgrd10m'
        self.prate_name = 'pratesfc'
        self.dlwrf_name = 'dlwrfsfc'
        self.dswrf_name = 'dswrfsfc'    
        self.air = None
        self.prc = None
        self.rad = None

    def write(
        self,
        outdir,
        start_date: datetime = None,
        rnday: Union[float, timedelta] = 4,
        air: bool = True,
        prc: bool = True,
        rad: bool = True,
        bbox: Bbox = None,
        overwrite: bool=False,
        output_interval: int = 1,
        tmpdir = None,
    ):
        self.start_date=start_date
        self.rnday=rnday
        self.end_date=self.start_date+timedelta(self.rnday + 1)

        dates = {_: None for _ in np.arange(
            self.start_date,
            self.end_date,
            np.timedelta64(1, 'D'),
            dtype='datetime64')}

        #write sflux_inputs.txt
        with open(f'{outdir}/sflux_inputs.txt', 'w') as f:
            f.write('&sflux_inputs\n/\n')

        logger.info('Start downloading ERA5')
        self.inventory = ERA5DataInventory(
            self.start_date,
            self.rnday,
            bbox,
            tmpdir = tmpdir
        )

        logger.info('Finished downloading ERA5')
    
        nx_grid, ny_grid = self.inventory.xy_grid()

        ds=Dataset(self.inventory.files[0])
        time1=ds['time']
        times=nc4.num2date(time1,units=time1.units,only_use_cftime_datetimes=False)

        for iday, date in enumerate(dates):
            put_sflux_fields(iday, date, times, ds, nx_grid, ny_grid, air=air, rad=rad, prc=prc, output_interval=output_interval, OUTDIR=outdir)




from datetime import datetime, timedelta
import logging
import pathlib
import tempfile
from time import time
import glob
import multiprocessing as mp

import boto3
from botocore import UNSIGNED
from botocore.config import Config
import numpy as np
import xarray as xr
import pandas as pd

from pyschism.dates import nearest_cycle

logger = logging.getLogger(__name__)

class AWSGrib2Inventory:

    def __init__(
            self,
            start_date: datetime = None,
            record = 1,
            pscr = None,
            product='atmos',
    ):
        """
        This will download the GFS data.
        """
        self.start_date = nearest_cycle() if start_date is None else start_date
        self.record = record
        self.pscr = pscr
        self.product = product
        self.forecast_cycle = self.start_date
        self.cycle=self.forecast_cycle.hour

        paginator=self.s3.get_paginator('list_objects_v2')
        pages=paginator.paginate(Bucket=self.bucket, 
                Prefix=f'gfs.{self.forecast_cycle.strftime("%Y%m%d")}'
                       f'/{self.cycle:02d}/{self.product}/')

        data=[]
        for page in pages:
            for obj in page['Contents']:
                data.append(obj) 

        file_metadata = list(sorted([
            _['Key'] for _ in data if '0p25' in _['Key'] and 'pgrb2' in _['Key'] and not 'goessim' in _['Key'] and not 'anl' in _['Key'] and not 'idx' in _['Key']]))

        for key in file_metadata[1:self.record*24+1]:
            key2 = key + '.grib2'
            filename = pathlib.Path(self.tmpdir) / key2
            filename.parent.mkdir(parents=True, exist_ok=True)

            with open(filename, 'wb') as f:
                logger.info(f'Downloading file {key}, ')
                try:
                    self.s3.download_fileobj(self.bucket, key, f)
                except:
                    logger.info(f'file {key} is not available')

    @property
    def bucket(self):
        return 'noaa-gfs-bdp-pds'

    @property
    def s3(self):
        try:
            return self._s3
        except AttributeError:
            self._s3 = boto3.client(
                's3', config=Config(signature_version=UNSIGNED))
            return self._s3

    @property
    def tmpdir(self):
        if not hasattr(self, "_tmpdir"):
            self._tmpdir = tempfile.TemporaryDirectory(dir=self.pscr)
        return pathlib.Path(self._tmpdir.name)

    @property
    def files(self):
        grbfiles=glob.glob(f'{self.tmpdir}/gfs.{self.forecast_cycle.strftime("%Y%m%d")}/{self.cycle:02d}/{self.product}/gfs.t{self.cycle:02d}z.pgrb2.0p25.f*.grib2')
        grbfiles.sort()
        return grbfiles

class GFS:
    def __init__(self, start_date=None, rnday=None, pscr=None, record=1, bbox=None, outdir=None): 
        start_date = nearest_cycle() if start_date is None else start_date 
        logger.info(f'start_date is {start_date}')
        self.bbox = bbox

        if outdir is None:
            self.path = pathlib.Path(start_date.strftime("%Y%m%d"))
            self.path.mkdir(parents=True, exist_ok=True)
        else:
            self.path = outdir

        end_date = start_date + timedelta(days=rnday)
        
        datevector = np.arange(start_date, end_date, np.timedelta64(1, 'D'),
                dtype='datetime64')
        datevector = pd.to_datetime(datevector)

        npool = len(datevector) if len(datevector) < mp.cpu_count()/2 else mp.cpu_count()/2
        logger.info(f'npool is {npool}')
        pool = mp.Pool(int(npool))

        pool.starmap(self.gen_sflux, [(date, record, pscr) for date in datevector])

        pool.close()

    def gen_sflux(self, date, record, pscr):

        inventory = AWSGrib2Inventory(date, record, pscr)
        grbfiles = inventory.files
        cycle = date.hour

        
        prate = []
        dlwrf = []
        dswrf = []
        stmp = []
        spfh = []
        uwind = []
        vwind = []
        prmsl = []

        Vars = {
            'group1': {'sh2': ['174096', spfh], 't2m': ['167', stmp],
            'u10': ['165', uwind], 'v10': ['166', vwind]},
            'group2': {'prmsl': ['meanSea', prmsl]},
            'group3': {'prate': ['surface', prate]},
            'group4': {'dlwrf': ['surface', dlwrf], 'dswrf': ['surface', dswrf]}
        }

        #Get lon/lat
        lon, lat, idx_ymin, idx_ymax, idx_xmin, idx_xmax = self.modified_latlon(grbfiles[0])

        for ifile, file in enumerate(grbfiles):
            logger.info(f'file {ifile} is {file}')

            for key, value in Vars.items():
                if key == 'group1':
                    for key2, value2 in value.items():
                        ds = xr.open_dataset(file,
                                engine='cfgrib',
                                backend_kwargs=dict(filter_by_keys={'paramId':int(value2[0])}))
                        tmp = ds[key2][idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                        value2[1].append(tmp[::-1,:])
                        ds.close()

                elif key == 'group2':
                    ds = xr.open_dataset(file,
                            engine='cfgrib',
                            backend_kwargs=dict(filter_by_keys={'typeOfLevel':'meanSea'}))
                    for key2, value2 in value.items():
                        tmp = ds[key2][idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                        value2[1].append(tmp[::-1,:])
                    ds.close()

                elif key == 'group3':
                    ds = xr.open_dataset(file,
                            engine='cfgrib',
                            backend_kwargs=dict(filter_by_keys={'stepType': 'instant','typeOfLevel': 'surface'}))
                    for key2, value2 in value.items():
                        tmp = ds[key2][idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                        value2[1].append(tmp[::-1, :])
                    ds.close()

                else:
                    ds = xr.open_dataset(file,
                            engine='cfgrib',
                            backend_kwargs=dict(filter_by_keys={'stepType': 'avg','typeOfLevel': 'surface'}))
                    for key2, value2 in value.items():
                        tmp = ds[key2][idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                        value2[1].append(tmp[::-1, :])
                    ds.close()

        fout = xr.Dataset({'stmp': (['time', 'ny_grid', 'nx_grid'], np.array(stmp)),
                'spfh': (['time', 'ny_grid', 'nx_grid'], np.array(spfh)),
                'uwind': (['time', 'ny_grid', 'nx_grid'], np.array(uwind)),
                'vwind': (['time', 'ny_grid', 'nx_grid'], np.array(vwind)),
                'prmsl': (['time', 'ny_grid', 'nx_grid'], np.array(prmsl)),
                'prate': (['time', 'ny_grid', 'nx_grid'], np.array(prate)),
                'dlwrf': (['time', 'ny_grid', 'nx_grid'], np.array(dlwrf)),
                'dswrf': (['time', 'ny_grid', 'nx_grid'], np.array(dswrf)),
                },
                coords={
                    'time': np.round(np.arange(1, len(grbfiles)+1)/24, 4).astype('float32'),
                    'lon': (['ny_grid', 'nx_grid'], lon),
                    'lat': (['ny_grid', 'nx_grid'], lat)
                })

        bdate = date.strftime('%Y %m %d %H').split(' ')
        bdate = [int(q) for q in bdate[:4]] + [0]

        fout.time.attrs = { 
            'long_name': 'Time',
            'standard_name': 'time',
            'base_date': bdate,
            'units': f"days since {date.year}-{date.month}-{date.day} {cycle:02d}:00 UTC"
        }

        fout.lat.attrs = {
            'units': 'degrees_north',
            'long_name': 'Latitude',
            'standard_name': 'latitude',
        }

        fout.lon.attrs = {
            'units': 'degrees_east',
            'long_name': 'Longitude',
            'standard_name': 'longitude',
        }

        fout.uwind.attrs={
            'units': 'm/s',
            'long_name': '10m_above_ground/UGRD',
            'standard_name':'eastward_wind'
        }

        fout.vwind.attrs={
            'units': 'm/s',
            'long_name': '10m_above_ground/VGRD',
            'standard_name':'northward_wind'
        }

        
        fout.spfh.attrs={
            'units': 'kg kg-1',
            'long_name': '2m_above_ground/Specific Humidity',
            'standard_name':'specific_humidity'
        }

        fout.prmsl.attrs = {
            'units': 'Pa',
            'long_name': 'Pressure reduced to MSL',
            'standard_name': 'air_pressure_at_sea_level'
        }

        fout.stmp.attrs={
            'units': 'K',
            'long_name': '2m_above_ground/Temperature',
        }

        fout.prate.attrs={
            'units': 'kg m-2 s-1',
            'long_name': 'Precipitation rate'
        }

        fout.dlwrf.attrs = {
            'units': 'W m-2',
            'long_name': 'Downward short-wave radiation flux'
        }

        fout.dswrf.attrs = {
            'units': 'W m-2',
            'long_name': 'Downward long-wave radiation flux'
        }
                         
        fout.to_netcdf(self.path / f'gfs_{date.strftime("%Y%m%d")}{cycle:02d}.nc','w', 'NETCDF3_CLASSIC', unlimited_dims='time')


    def modified_latlon(self, grbfile):
        xmin, xmax, ymin, ymax = self.bbox.xmin, self.bbox.xmax, self.bbox.ymin, self.bbox.ymax
        xmin = xmin + 360 if xmin < 0 else xmin
        xmax = xmax + 360 if xmax < 0 else xmax

        ds=xr.open_dataset(grbfile,
                engine='cfgrib',
                backend_kwargs=dict(filter_by_keys={'stepType': 'instant','typeOfLevel': 'surface'}))
        lon=ds.longitude.astype('float32')
        lat=ds.latitude.astype('float32')
        lon_idxs=np.where((lon.values >= xmin-1.0) & (lon.values <= xmax+1.0))[0]
        lat_idxs=np.where((lat.values >= ymin-1.0) & (lat.values <= ymax+1.0))[0]
        idx_ymin = lat_idxs[0]
        idx_ymax = lat_idxs[-1]
        idx_xmin = lon_idxs[0]
        idx_xmax = lon_idxs[-1]
        lon2 = lon[lon_idxs]
        lat2 = lat[lat_idxs]
        idxs = np.where(lon2 > 180)
        lon2[idxs] -= 360
        logger.info(f'idx_ymin is {idx_ymin}, idx_ymax is {idx_ymax}, idx_xmin is {idx_xmin}, idx_xmax is {idx_xmax}')
        #make sure lat is in ascending order
        nx_grid, ny_grid=np.meshgrid(lon2, lat2[::-1])

        ds.close()

        return nx_grid, ny_grid, idx_ymin, idx_ymax, idx_xmin, idx_xmax



import logging
import os
import pathlib

import appdirs
from netCDF4 import Dataset
import numpy as np
from scipy.interpolate import griddata
from scipy.interpolate.fitpack2 import RectBivariateSpline

from pyschism.forcing.bctides.base import TidalDataProvider


logger = logging.getLogger(__name__)

TPXO_ELEVATION = 'h_tpxo9.v1.nc'
TPXO_VELOCITY = 'u_tpxo9.v1.nc'


def raise_missing_file(fpath, fname):
    raise FileNotFoundError('\n'.join([
        f'No TPXO file found at "{fpath}".',
        'New users will need to register and request a copy of '
        f'the TPXO9 NetCDF file (specifically `{fname}`) '
        'from the authors at https://www.tpxo.net.',
        'Once you obtain `h_tpxo9.v1.nc`, you can follow one of the '
        'following options: ',
        f'1) copy or symlink the file to "{fpath}"',
        f'2) set the environment variable `{fname}` to point'
        ' to the file',
    ]))


class TPXO(TidalDataProvider):

    def __init__(self, h_file=None, u_file=None):
        self._h_file = h_file
        self._u_file = u_file

    def get_elevation(self, constituent, vertices):
        logger.info('Querying TPXO for elevation constituent '
                    f'{constituent}.')
        amp = self._get_interpolation(
            'elevation', 'ha', constituent, vertices)
        phase = self._get_interpolation(
            'elevation', 'hp', constituent, vertices)
        return amp, phase

    def get_velocity(self, constituent, vertices):
        logger.info('Querying TPXO for velocity constituent '
                    f'{constituent}.')
        uamp = self._get_interpolation(
            'velocity', 'ua', constituent, vertices) / 100.
        uphase = self._get_interpolation(
            'velocity', 'up', constituent, vertices)
        vamp = self._get_interpolation(
            'velocity', 'va', constituent, vertices) / 100.
        vphase = self._get_interpolation(
            'velocity', 'vp', constituent, vertices)
        return uamp, uphase, vamp, vphase

    @property
    def constituents(self):
        return ['M2', 'S2', 'N2', 'K2', 'K1', 'O1', 'P1',
                'Q1', 'Mm', 'Mf', 'M4', 'MN4', 'MS4', '2N2', 'S1']
        if not hasattr(self, '_constituents'):
            self._constituents = [
                c.capitalize() for c in self.h['con'][:].astype(
                    '|S1').tobytes().decode('utf-8').split()]
        return self._constituents

    @property
    def lon_z(self) -> np.ndarray:
        return self.h['lon_z'][:, 0].data

    @property
    def lat_z(self) -> np.ndarray:
        return self.h['lat_z'][0, :].data

    @property
    def lon_u(self) -> np.ndarray:
        return self.uv['lon_u'][:, 0].data

    @property
    def lat_u(self) -> np.ndarray:
        return self.uv['lat_u'][0, :].data

    @property
    def lon_v(self) -> np.ndarray:
        return self.uv['lon_v'][:, 0].data

    @property
    def lat_v(self) -> np.ndarray:
        return self.uv['lat_v'][0, :].data

    @property
    def h(self):
        if not hasattr(self, '_h'):
            if self._h_file is None:
                self._h_file = os.getenv('TPXO_ELEVATION')
                if self._h_file is None:
                    self._h_file = pathlib.Path(
                        appdirs.user_data_dir('tpxo')) / TPXO_ELEVATION
            if not self._h_file.exists():
                raise_missing_file(self._h_file, TPXO_ELEVATION)
            self._h = Dataset(self._h_file)
        return self._h

    @property
    def uv(self):
        if not hasattr(self, '_uv'):
            if self._u_file is None:
                self._u_file = os.getenv('TPXO_VELOCITY')
                if self._u_file is None:
                    self._u_file = pathlib.Path(
                        appdirs.user_data_dir('tpxo')) / TPXO_VELOCITY
            if not self._u_file.exists():
                raise_missing_file(self._u_file, TPXO_VELOCITY)
            self._uv = Dataset(self._u_file)
        return self._uv

    def _get_interpolation(self, phys_var, ncvar, constituent, vertices):
        lower_c = [c.lower() for c in self.constituents]
        if phys_var == 'elevation':
            ncarray = self.h
            self.x = self.lon_z
            self.y = self.lat_z
        elif ncvar == 'ua' or ncvar == 'up':
            ncarray = self.uv
            self.x = self.lon_u
            self.y = self.lat_u
        elif ncvar == 'va' or ncvar == 'vp':
            ncarray = self.uv
            self.x = self.lon_v
            self.y = self.lat_v

        zi = ncarray[ncvar][
            lower_c.index(constituent.lower()), :, :]
        xo = np.asarray(
            [x + 360. if x < 0. else x for x in vertices[:, 0]]).flatten()
        yo = vertices[:, 1].flatten()
        xi, yi = np.meshgrid(self.x, self.y, indexing='ij')
        xi = xi.flatten()
        yi = yi.flatten()
        zi = zi.flatten()
        dx = np.mean(np.diff(self.x))
        dy = np.mean(np.diff(self.y))
        # buffer the bbox by 2 difference units
        mask1 = np.logical_and(
            np.logical_and(
                xi >= np.min(xo) - 2 * dx,
                xi <= np.max(xo) + 2 * dx
            ),
            np.logical_and(
                yi >= np.min(yo) - 2 * dy,
                yi <= np.max(yo) + 2 * dy
            )
        )
        # remove junk values from input array
        mask2 = np.ma.masked_where(zi != 0., zi)
        iidx = np.where(np.logical_and(mask1, mask2))
        values = griddata(
            (xi[iidx], yi[iidx]),
            zi[iidx],
            (xo, yo),
            method='linear',
            fill_value=np.nan,
        )
        nan_idxs = np.where(np.isnan(values))
        values[nan_idxs] = griddata(
            (xi[iidx], yi[iidx]),
            zi[iidx],
            (xo[nan_idxs], yo[nan_idxs]),
            method='nearest',
        )
        return values



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
import logging
from typing import Dict, Union

from matplotlib.transforms import Bbox
from netCDF4 import Dataset
import numpy as np
from scipy.interpolate import griddata
import tqdm
import tqdm_logging_wrapper

from pyschism import dates
from pyschism.forcing.base import ModelForcing


logger = logging.getLogger(__name__)


class HycomComponent(ABC):
    '''Base class for HYCOM component API'''

    @abstractmethod
    def get_datasets(
            self,
            start_date,
            run_days,
            output_interval=None
    ) -> Dict[datetime, Dataset]:
        '''Returns a dictionary of relevant datasets.'''

    @abstractmethod
    def put_boundary_ncdata(*args, **kwargs):
        '''Interpolates HYCOM data to boundary.'''

    @property
    @abstractmethod
    def ncvar(self) -> Union[str, tuple]:
        '''Name of the variable or variables in the HYCOM dataset.'''

    def _modified_bbox(self, dataset, bbox=None):
        # dataset = list(self.datasets.values())[-1]
        if bbox is None:
            return Bbox.from_extents(
                dataset['lon'][:].min(),
                dataset['lat'][:].min(),
                dataset['lon'][:].max(),
                dataset['lat'][:].max()
            )
        else:
            xmin = bbox.xmin + 360. if not (
                bbox.xmin >= dataset['lon'][:].min()
                and bbox.xmin < 180.) else bbox.xmin

            xmax = bbox.xmax + 360. if not (
                bbox.xmax >= dataset['lon'][:].min()
                and bbox.xmax < 180.) else bbox.xmax

            return Bbox.from_extents(
                np.min([xmin, xmax]),
                bbox.ymin,
                np.max([xmin, xmax]),
                bbox.ymax
            )

    def _modified_bbox_indexes(
            self,
            bbox,
            dataset,
            pixel_buffer=0
    ):
        # dataset = list(self.datasets.values())[-1]
        lat_idxs = np.where((dataset['lat'][:] >= bbox.ymin)
                            & (dataset['lat'][:] <= bbox.ymax))[0]
        lon_idxs = np.where((dataset['lon'][:] >= bbox.xmin)
                            & (dataset['lon'][:] <= bbox.xmax))[0]
        lon_idxs = lon_idxs.tolist()
        lat_idxs = lat_idxs.tolist()
        for i in range(pixel_buffer):
            lon_idxs.insert(0, lon_idxs[0] - 1)
            lon_idxs.append(lon_idxs[-1] + 1)
            lat_idxs.insert(0, lat_idxs[0] - 1)
            lat_idxs.append(lat_idxs[-1] + 1)
        return lon_idxs, lat_idxs


class Hycom(ModelForcing):
    
    @property
    @abstractmethod
    def elevation(self) -> HycomComponent:
        '''Elevation component API of HYCOM dataset.'''

    @property
    @abstractmethod
    def velocity(self) -> HycomComponent:
        '''Elevation component API of HYCOM dataset.'''

    @property
    @abstractmethod
    def temperature(self) -> HycomComponent:
        '''Temperature component API of HYCOM dataset.'''

    @property
    @abstractmethod
    def salinity(self) -> HycomComponent:
        '''Temperature component API of HYCOM dataset.'''



from pyschism.mesh.base import Gr3
from pyschism.mesh.hgrid import Hgrid
from pyschism.mesh.vgrid import Vgrid
from pyschism.mesh.fgrid import (
    Fgrid, RoughnessLength, ManningsN, DragCoefficient)

__all__ = ["Gr3", "Hgrid", "Vgrid", "Fgrid", "RoughnessLength", "ManningsN",
           "DragCoefficient"]



#! /usr/bin/env python
import argparse
# from datetime import datetime
import logging
import os
import pathlib
# import shutil
import tarfile
import tempfile
import unittest
import urllib.request
from datetime import datetime, timedelta

from pyschism.cmd.bctides import BctidesCli
from pyschism.driver import ModelConfig
from pyschism.mesh import Hgrid


logging.basicConfig(level=logging.INFO, force=True)


class ModelConfigurationTestCase(unittest.TestCase):

    def setUp(self):
        hgrid2D = os.getenv('NWM_TEST_MESH')
        if hgrid2D is None:
            data_directory = pathlib.Path(__file__).parent.absolute() / 'data'
            hgrid2D = data_directory / "GulfStreamDevel/hgrid.gr3"
            url = "https://www.dropbox.com/s/mjaxaqeggy721um/"
            url += "Gulf_Stream_develop.tar.gz?dl=1"
            g = urllib.request.urlopen(url)
            tmpfile = tempfile.NamedTemporaryFile()
            with open(tmpfile.name, 'b+w') as f:
                f.write(g.read())
            with tarfile.open(tmpfile.name, "r:gz") as tar:
                tar.extractall(data_directory / "GulfStreamDevel")
        self.hgrid2D = hgrid2D

    #def _test_bctides_cli_2d(self):
    #    import tempfile
    #    tmpdir = tempfile.TemporaryDirectory()
    #    parser = argparse.ArgumentParser()
    #    add_bctides(parser.add_subparsers(dest='mode'))
    #    args = parser.parse_args(
    #        [
    #            'bctides',
    #            f'{self.hgrid2D}',  # hgrid path
    #            '--run-days=5',  # rndays
    #            '--hgrid-crs=epsg:4326',
    #            f'-o={tmpdir.name}',
    #            '--overwrite'
    #        ])
    #    BctidesCli(args)
    #    # with open(tmpdir.name + '/bctides.in') as f:
    #    #     for line in f:
    #    #         print(line, end='')

    #def _test_bctides_cli_2d_w_velo(self):
    #    import tempfile
    #    tmpdir = tempfile.TemporaryDirectory()
    #    parser = argparse.ArgumentParser()
    #    add_bctides(parser.add_subparsers(dest='mode'))
    #    args = parser.parse_args(
    #        [
    #            'bctides',
    #            f'{self.hgrid2D}',  # hgrid path
    #            '--run-days=5',  # rndays,
    #            '--include-velocity',
    #            '--hgrid-crs=epsg:4326',
    #            f'-o={tmpdir.name}',
    #            '--overwrite'
    #        ])
    #    BctidesCli(args)
    #    # with open(tmpdir.name + '/bctides.in') as f:
    #    #     for line in f:
    #    #         print(line, end='')

    #def test_bctides_cli_2d_w_elev2d(self):
    #    import tempfile
    #    tmpdir = tempfile.TemporaryDirectory()
    #    parser = argparse.ArgumentParser()
    #    add_bctides(parser.add_subparsers(dest='mode'))
    #    args = parser.parse_args(
    #        [
    #            'bctides',
    #            f'{self.hgrid2D}',  # hgrid path
    #            '--run-days=5',  # rndays,
    #            '--iettype-5',
    #            # '--ifltype-5',
    #            '--hgrid-crs=epsg:4326',
    #            f'-o={tmpdir.name}',
    #            '--overwrite'
    #        ])
    #    BctidesCli(args)

    #def _test_bctides_cli_3d(self):
    #    import tempfile
    #    tmpdir = tempfile.TemporaryDirectory()
    #    parser = argparse.ArgumentParser()
    #    add_bctides(parser.add_subparsers(dest='mode'))
    #    args = parser.parse_args(
    #        [
    #            'bctides',
    #            f'{self.hgrid}',  # hgrid path
    #            '--run-days=5',  # rndays
    #            '--hgrid-crs=epsg:4326',
    #            f'-o={tmpdir.name}',
    #            '--overwrite'
    #        ])
    #    BctidesCli(args)

    def test_case1_setup(self):
        # Inputs: Modify As Needed!
        start_date = "2018091000" #in YYYYMMDDHH format
        end_date = "2018091600" #in YYYYMMDDHH format

        dt = timedelta(seconds=150.)
        nspool = timedelta(minutes=20.)

        start_date = datetime.strptime(start_date, "%Y%m%d%H")
        end_date = datetime.strptime(end_date, "%Y%m%d%H")
        rnday = end_date - start_date

        hgrid = Hgrid.open(
            'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll',
            crs=4326
        )

        config = ModelConfig(
            hgrid=hgrid,
            flags=[[3, 3, 0, 0]],
            constituents = 'major',
            database = 'tpxo',
        )

        coldstart = config.coldstart(
            start_date=start_date,
            end_date=start_date + rnday,
            timestep=dt,
            nspool=timedelta(hours=1),
            elev=True,
            dahv=True,
        )

        with tempfile.TemporaryDirectory() as dn:
            tmpdir = pathlib.Path(dn)
            coldstart.write(tmpdir, overwrite=True)

    def test_case2_setup(self):
        # Inputs: Modify As Needed!
        start_date = "2018091000" #in YYYYMMDDHH format
        end_date = "2018091600" #in YYYYMMDDHH format

        dt = timedelta(seconds=150.)
        nspool = timedelta(minutes=20.)

        start_date = datetime.strptime(start_date, "%Y%m%d%H")
        end_date = datetime.strptime(end_date, "%Y%m%d%H")
        rnday = end_date - start_date

        hgrid = Hgrid.open(
            'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll',
            crs=4326
        )

        config = ModelConfig(
            hgrid=hgrid,
            flags=[[5, 5, 0, 0]],
            constituents = 'major',
            database = 'tpxo',
        )

        coldstart = config.coldstart(
            start_date=start_date,
            end_date=start_date + rnday,
            timestep=dt,
            nspool=timedelta(hours=1),
            elev=True,
            dahv=True,
        )

        with tempfile.TemporaryDirectory() as dn:
            tmpdir = pathlib.Path(dn)
            coldstart.write(tmpdir, overwrite=True)

    def test_case3_setup(self):
        # Inputs: Modify As Needed!
        start_date = "2018091000" #in YYYYMMDDHH format
        end_date = "2018091600" #in YYYYMMDDHH format

        dt = timedelta(seconds=150.)
        nspool = timedelta(minutes=20.)

        start_date = datetime.strptime(start_date, "%Y%m%d%H")
        end_date = datetime.strptime(end_date, "%Y%m%d%H")
        rnday = end_date - start_date

        hgrid = Hgrid.open(
            'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll',
            crs=4326
        )

        config = ModelConfig(
            hgrid=hgrid,
            flags=[[5, 5, 2, 2]], #constant temperature and salinity
            constituents = 'major',
            database = 'tpxo',
            tthconst = [10.0],   #T value
            sthconst = [0.0],    #S value
            tobc = [0.1],   #nuding factor for T
            sobc = [0.1],   #nuding factor for S
        )

        coldstart = config.coldstart(
            start_date=start_date,
            end_date=start_date + rnday,
            timestep=dt,
            nspool=timedelta(hours=1),
            elev=True,
            dahv=True,
        )

        with tempfile.TemporaryDirectory() as dn:
            tmpdir = pathlib.Path(dn)
            coldstart.write(tmpdir, overwrite=True)

    def test_case4_setup(self):
        # Inputs: Modify As Needed!
        start_date = "2018091000" #in YYYYMMDDHH format
        end_date = "2018091600" #in YYYYMMDDHH format

        dt = timedelta(seconds=150.)
        nspool = timedelta(minutes=20.)

        start_date = datetime.strptime(start_date, "%Y%m%d%H")
        end_date = datetime.strptime(end_date, "%Y%m%d%H")
        rnday = end_date - start_date

        hgrid = Hgrid.open(
            'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll',
            crs=4326
        )

        config = ModelConfig(
            hgrid=hgrid,
            flags=[[5, 5, 4, 4]], #constant temperature and salinity
            constituents = 'major',
            database = 'tpxo',
            tobc = [0.1],   #nuding factor for T
            sobc = [0.1],   #nuding factor for S
        )

        coldstart = config.coldstart(
            start_date=start_date,
            end_date=start_date + rnday,
            timestep=dt,
            nspool=timedelta(hours=1),
            elev=True,
            dahv=True,
        )

        with tempfile.TemporaryDirectory() as dn:
            tmpdir = pathlib.Path(dn)
            coldstart.write(tmpdir, overwrite=True)

if __name__ == '__main__':
    unittest.main()



from time import time
import numpy as np
from glob import glob 
from scipy import spatial
import xarray as xr
import pandas as pd


from pyschism.mesh import Hgrid
from pyschism.forcing.source_sink.sflux2source.SourceSinkIn import source_sink


def nearest_neighbour(points_a, points_b):
    tree = spatial.cKDTree(points_b)
    return tree.query(points_a)[1]

def Sflux_var_at_points(sflux_dir, sflux_set='air_1', points=None, var_name=None, nt=None):
    '''
    extract variable values at specified points from SCHISM-format sflux files
    '''

    sflux_files = sorted(glob(f'{sflux_dir}/sflux_{sflux_set}.*.nc'))
    # read the first sflux file
    sflux = xr.open_dataset(sflux_files[0])
    if nt is None:
        nt = len(sflux['time'].data)
    sflux.close()

    idx = nearest_neighbour(np.c_[points[:, 0], points[:, 1]], np.c_[sflux.lon.data.ravel(), sflux.lat.data.ravel()])
    sflux_files = sorted(glob(f'{sflux_dir}/sflux_{sflux_set}.*.nc'))
    nt_all = nt * len(sflux_files)
    npts = len(idx)

    var_at_pts = np.empty((nt_all, npts), dtype=float) * np.nan
    time = np.empty((nt_all), dtype='datetime64[ns]')

    for i, sflux_file in enumerate(sflux_files):
        sflux = xr.open_dataset(sflux_file)
        var = sflux[var_name].data
        var = var.reshape(*var.shape[:-2], -1)
        var_at_pts[i*nt:(i+1)*nt, :] = var[:nt, idx]
        time[i*nt:(i+1)*nt] = sflux.time.data[:nt]
        print(f'processing {sflux_file}')
        sflux.close()

    return [time, var_at_pts]


if __name__ == "__main__":
    rundir = '/sciclone/schism10/lcui01/schism20/ICOGS/ICOGS2D/GA_RUN02'
    model_start_time_str = "2018-08-01 00:00:00"

    # read hgrid and get element info
    hgrid = Hgrid.open(f'{rundir}/hgrid.gr3', crs='epsg:4326')
    lon_ctr, lat_ctr, dp_ctr = hgrid.elements.compute_centroid()
    #get element area (crs=26918 to project the coordinate)
    area = hgrid.elements.get_areas(26918)

    ## read the original source/sink files
    original_source_sink = source_sink(source_dir=rundir, start_time_str=model_start_time_str)

    # make another set of source/sink files based on sflux
    # read variable values from sflux
    var_at_ele_center = Sflux_var_at_points(f'{rundir}/sflux/', sflux_set='prc_1', points=np.c_[lon_ctr, lat_ctr], var_name='prate')
    # read time info from sflux
    times = pd.to_datetime(var_at_ele_center[0].astype(str))
    start_time_str = times[0].strftime("%Y-%m-%d %H:%M:%S")
    timedeltas = (times-times[0]).total_seconds().astype(int)
    nt = times.shape[0]
    # calculate new sources for all elements: prate x area
    sources_sflux = np.zeros((nt, hgrid.elements.__len__()), dtype=float)
    for it in range(nt):  # for each time
        sources_sflux[it, :] = 1e-3 * var_at_ele_center[1][it] * area  # prate is in kg/m^2/s, i.e., 1e-3 m3/m2/s

    ## make new source sink 
    added_ss = source_sink(
        source_dir=None,
        source_eles=(np.array(range(hgrid.elements.__len__()))+1).tolist(),
        sink_eles=[1],  # dummy
        start_time_str=start_time_str,
        timedeltas=timedeltas,
        vsource_data=sources_sflux
    )
    added_ss.update_vars()

    ## add original source and sflux source
    print('Adding two sources:')
    t0 = time()
    total_ss = original_source_sink + added_ss
    if np.isnan(total_ss.vsource.data.astype(float)).any():
        raise Exception('nan found in sources')
    if np.isnan(total_ss.vsink.data.astype(float)).any():
        raise Exception('nan found in sinks')

    print(f'It took {(time()-t0)/60} minutes to add two sources')
    
    ## write
    total_ss.nc_writer(f'{rundir}/Orignial+sflux_source_sink/') 
